// Copyright (C) 2002 Michael Ringgaard.
// strftime ecrit pour le domaine public par jim nutt
// VF et adaptation pour microcontroleurs PIC18 : Christian Dupaty

// BIBLIOTHEQUE TIME.C pour PIC18 compatible CANSI 
// compilateur MCC18

#include "time.h"

#define YEAR0                   1900
#define EPOCH_YR                1970
#define SECS_DAY                (86400L)
#define LEAPYEAR(year)          (!((year) % 4) && (((year) % 100) || !((year) % 400)))
#define YEARSIZE(year)          (LEAPYEAR(year) ? 366 : 365)
#define FIRSTSUNDAY(timp)       (((timp)->tm_yday - (timp)->tm_wday + 420) % 7)
#define FIRSTDAYOF(timp)        (((timp)->tm_wday - (timp)->tm_yday + 420) % 7)

#define TIME_MAX                2147483647L

rom const char *aday[] = {
    "Dim", "Lun", "Mar", "Mer", "Jeu", "Ven", "Sam"
};

rom const char *day[] = {
    "Dimanche", "Lundi", "Mardi", "Mercredi",
    "Jeudi", "Vendredi", "Samedi"
};

rom const char *amonth[] = {
    "Jan", "Fev", "Mar", "Avr", "Mai", "Juin",
    "Jui", "Aou", "Sep", "Oct", "Nov", "Dec"
};

rom const char *month[] = {
    "Janvier", "Fevrier", "Mars", "Avril", "Mai", "Juin",
    "Juillet", "Aout", "Septembre", "Octobre", "Novembre", "Decembre"
};

rom const int _ytab[2][12] = 
{
  {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
  {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};

#define ASCBUFSIZE 30

extern clock_t timerCPU; // compteur de secondes CPU
extern time_t timerCAL; //  compteur calendaire

struct tm tmtampon;
struct tm *tmbuf;
unsigned char ascbuf[ASCBUFSIZE];

int _daylight = 0;                  // different de 0 si la sauvegarde du decalage horaire local est utilisee
long _dstbias = 0;                  // decalage entre l'heure solaire et locale
long _timezone = 0;                 // diffrence entre le temps local et GMT
char *_tzname[2] = {"GMT", "PAR"};  // Fuseaux horaires standards et local 

struct tm *gmtime_r(const time_t *timer, struct tm *tmbuf)
{
  time_t time = *timer;
  unsigned long dayclock, dayno;
  int year = EPOCH_YR;

  dayclock = (unsigned long) time % SECS_DAY;
  dayno = (unsigned long) time / SECS_DAY;

  tmbuf->tm_sec = dayclock % 60;
  tmbuf->tm_min = (dayclock % 3600) / 60;
  tmbuf->tm_hour = dayclock / 3600;
  tmbuf->tm_wday = (dayno + 4) % 7; // Day 0 was a thursday
  while (dayno >= (unsigned long) YEARSIZE(year)) 
  {
    dayno -= YEARSIZE(year);
    year++;
  }
  tmbuf->tm_year = year - YEAR0;
  tmbuf->tm_yday = dayno;
  tmbuf->tm_mon = 0;
  while (dayno >= (unsigned long) _ytab[LEAPYEAR(year)][tmbuf->tm_mon]) 
  {
    dayno -= _ytab[LEAPYEAR(year)][tmbuf->tm_mon];
    tmbuf->tm_mon++;
  }
  tmbuf->tm_mday = dayno + 1;
  tmbuf->tm_isdst = 0;

  return tmbuf;
}

struct tm *localtime_r(const time_t *timer, struct tm *tmbuf)
{
  time_t t;

  t = *timer - _timezone;
  return gmtime_r(&t, tmbuf);
}


struct tm *gmtime(const time_t *timer)
{
  return gmtime_r(timer, tmbuf);
}

struct tm *localtime(const time_t *timer)
{
  tmbuf=&tmtampon;	
  return localtime_r(timer, tmbuf);
}

time_t mktime(struct tm *tmbuf)
{
  long day, year;
  int tm_year;
  int yday, month;
  unsigned long seconds;
  int overflow;
  long dst;

  tmbuf->tm_min += tmbuf->tm_sec / 60;
  tmbuf->tm_sec %= 60;
  if (tmbuf->tm_sec < 0) 
  {
    tmbuf->tm_sec += 60;
    tmbuf->tm_min--;
  }
  tmbuf->tm_hour += tmbuf->tm_min / 60;
  tmbuf->tm_min = tmbuf->tm_min % 60;
  if (tmbuf->tm_min < 0) 
  {
    tmbuf->tm_min += 60;
    tmbuf->tm_hour--;
  }
  day = tmbuf->tm_hour / 24;
  tmbuf->tm_hour= tmbuf->tm_hour % 24;
  if (tmbuf->tm_hour < 0) 
  {
    tmbuf->tm_hour += 24;
    day--;
  }
  tmbuf->tm_year += tmbuf->tm_mon / 12;
  tmbuf->tm_mon %= 12;
  if (tmbuf->tm_mon < 0) 
  {
    tmbuf->tm_mon += 12;
    tmbuf->tm_year--;
  }
  day += (tmbuf->tm_mday - 1);
  while (day < 0) 
  {
    if(--tmbuf->tm_mon < 0) 
    {
      tmbuf->tm_year--;
      tmbuf->tm_mon = 11;
    }
    day += _ytab[LEAPYEAR(YEAR0 + tmbuf->tm_year)][tmbuf->tm_mon];
  }
  while (day >= _ytab[LEAPYEAR(YEAR0 + tmbuf->tm_year)][tmbuf->tm_mon]) 
  {
    day -= _ytab[LEAPYEAR(YEAR0 + tmbuf->tm_year)][tmbuf->tm_mon];
    if (++(tmbuf->tm_mon) == 12) 
    {
      tmbuf->tm_mon = 0;
      tmbuf->tm_year++;
    }
  }
  tmbuf->tm_mday = day + 1;
  year = EPOCH_YR;
  if (tmbuf->tm_year < year - YEAR0) return (time_t) -1;
  seconds = 0;
  day = 0;                      // Means days since day 0 now
  overflow = 0;

  // Assume that when day becomes negative, there will certainly
  // be overflow on seconds.
  // The check for overflow needs not to be done for leapyears
  // divisible by 400.
  // The code only works when year (1970) is not a leapyear.
  tm_year = tmbuf->tm_year + YEAR0;

  if (TIME_MAX / 365 < tm_year - year) overflow++;
  day = (tm_year - year) * 365;
  if (TIME_MAX - day < (tm_year - year) / 4 + 1) overflow++;
  day += (tm_year - year) / 4 + ((tm_year % 4) && tm_year % 4 < year % 4);
  day -= (tm_year - year) / 100 + ((tm_year % 100) && tm_year % 100 < year % 100);
  day += (tm_year - year) / 400 + ((tm_year % 400) && tm_year % 400 < year % 400);

  yday = month = 0;
  while (month < tmbuf->tm_mon)
  {
    yday += _ytab[LEAPYEAR(tm_year)][month];
    month++;
  }
  yday += (tmbuf->tm_mday - 1);
  if (day + yday < 0) overflow++;
  day += yday;

  tmbuf->tm_yday = yday;
  tmbuf->tm_wday = (day + 4) % 7;               // Day 0 was thursday (4)

  seconds = ((tmbuf->tm_hour * 60L) + tmbuf->tm_min) * 60L + tmbuf->tm_sec;

  if ((TIME_MAX - seconds) / SECS_DAY < day) overflow++;
  seconds += day * SECS_DAY;

  // Now adjust according to timezone and daylight saving time
  if (((_timezone > 0) && (TIME_MAX - _timezone < seconds))
      || ((_timezone < 0) && (seconds < -_timezone)))
          overflow++;
  seconds += _timezone;

  if (tmbuf->tm_isdst)
    dst = _dstbias;
  else 
    dst = 0;

  if (dst > seconds) overflow++;        // dst is always non-negative
  seconds -= dst;

  if (overflow) return (time_t) -1;

  if ((time_t) seconds != seconds) return (time_t) -1;
  return (time_t) seconds;
}

char *asctime(const struct tm *tp)
{
  strftime((char *)ascbuf, ASCBUFSIZE,(rom const char *)"%c\n", tp);
  return (char *)ascbuf ;
}

char *ctime(const time_t *tp)
{
  return asctime(localtime(tp));
}

char *_strdate(char *s)
{
  time_t now;

  time(&now);
  strftime(s, 9, "%D", localtime(&now));
  return s;
}

char *_strtime(char *s)
{
  time_t now;

  time(&now);
  strftime(s, 9, "%T", localtime(&now));
  return s;
}

void _tzset()
{
}

extern time_t timer;
time_t time(time_t *__timer)
{
	return (*__timer);
}

clock_t clock(void)
{
	return (timerCPU);
}

/**
 * strftime.c
 * version FR et adaptation PIC18 par Christian Dupaty 05-2006
 *
 * size_t strftime_(char *str,
 *                  size_t maxs,
 *                  const char *fmt,
 *                  const struct tm *t)
 *
 *      Cette fonction formate une chaine de carateres avec les memes parametres que sprintf
 *
 *          str est l'adresse du tableau de caracteres pour sprintf
 *
 *          maxs est la taille du tableau
 *
 *          fmt indique le format de sortie il est precede de %
 *
 *              %A      jour de la semaine (Lundi)
 *              %a      jour de la semaine abrege (Lun)
 *              %B      mois (Janvier)
 *              %b      mois abrege (Jan)
 *              %c      date et heure abrege
 *              %d      jour du moid (01-31)
 *              %H      heure (00-23)
 *              %I      heure (01-12)
 *              %j      annee (001-366)
 *              %M      minute (00-59)
 *              %m      mois (01-12)
 *              %p      tepms local  AM ou PM
 *              %S      secondes (00-59)
 *              %U      numero de semaine avec dimanche comme premier jour (00-53)
 *              %W      numero de semaine avec lundi comme premier jour (00-53)
 *              %w      numero du jour de la semaine (0-6, dimanche est 0)
 *              %X      Representation standard du temps 
 *              %x      Representation standard de la date 
 *              %Y      annee avec siecle (2006)
 *              %y      anne sans le siecle (00-99)
 *              %Z      nom du fuseau horaire local
 *              %%      signe pourcent
 *
 *
 *      strftime_() retourne le nombre de caracteres dans le tableau de sortie
 *
*/

size_t strftime(char *s, size_t maxs, rom const char *f, const struct tm *t)
{
      int w;
      char *q;

       q = s + maxs - 1;
      while ((*f != '\0'))
      {
            if (*f++ == '%')
            {
                  switch (*f++)
                  {
                  case '%' :
                        s = "%";
                        break;

                  case 'a' :
                        s = aday[t->tm_wday];
                        break;

                  case 'A' :
                        s = day[t->tm_wday];
                        break;

                  case 'b' :
                        s = amonth[t->tm_mon];
                        break;

                  case 'B' :
                        s = month[t->tm_mon];
                        break;

                  case 'c' :
                        sprintf(s,"%HS %d %HS %d - %2d:%2d:%2d",day[t->tm_wday], t->tm_mday,month[t->tm_mon],t->tm_year+1900,t->tm_hour, t->tm_min,t->tm_sec);
                        break;

                  case 'd' :
                        sprintf(s,"%2d",t->tm_mday);
                        break;

                  case 'H' :
                        sprintf(s,"%2d",t->tm_hour);
                        break;

                  case 'I' :
                        sprintf(s,"%2d",(t->tm_hour%12)?t->tm_hour%12:12);
                        break;

                  case 'j' :
                        sprintf(s,"%3d",t->tm_yday+1);
                        break;

                  case 'm' :
                        sprintf(s,"%2d",t->tm_mon+1);
                        break;

                  case 'M' :
                        sprintf(s,"%2d",t->tm_min);
                        break;

                  case 'p' :
                        s = (t->tm_hour>11)?"PM":"AM";
                        break;

                  case 'S' :
                        sprintf(s,"%2d",t->tm_sec);
                        break;

                  case 'U' :
                        w = t->tm_yday/7;
                        if (t->tm_yday%7 > t->tm_wday)
                              w++;
                        sprintf(s, "%2d", w);
                        break;

                  case 'W' :
                        w = t->tm_yday/7;
                        if (t->tm_yday%7 > (t->tm_wday+6)%7)
                              w++;
                        sprintf(s, "%2d", w);
                        break;

                  case 'w' :
                        sprintf(s,"%1d",t->tm_wday);
                        break;

                  case 'x' :
                        sprintf(s, "%HS %2d %HS %4d", aday[t->tm_wday], t->tm_mday,amonth[t->tm_mon], t->tm_year+1900);
                        break;

                  case 'X' :
                        sprintf(s, "%2d:%2d:%2d", t->tm_hour,t->tm_min, t->tm_sec);
                        break;

                  case 'y' :
                        sprintf(s,"%2d",t->tm_year%100);
                        break;

                  case 'Y' :
                        sprintf(s,"%4d",t->tm_year+1900);
                        break;

                  case 'Z' :
                        s = (t->tm_isdst && _tzname[1][0]) ?
                              _tzname[1] : _tzname[0];
                        break;

                  default:
                        s[0] = '%';     /* reconstruit le format */
                        s[1] = f[-1];
                        s[2] = '\0';
                        if (s[1] == 0)
                              f--;      
                  }
                 
            }
            else
            return 0;
        }

      return q - s;
}
